//-----------------------------------------------------------------------------
// The confidential and proprietary information contained in this file may
// only be used by a person authorised under and to the extent permitted
// by a subsisting licensing agreement from Arm Limited or its affiliates.
//
//            (C) COPYRIGHT 2012-2016 Arm Limited or its affiliates.
//                ALL RIGHTS RESERVED
//
// This entire notice must be reproduced on all copies of this file
// and copies of this file may only be made by a person if such person is
// permitted to do so under the terms of a subsisting license agreement
// from Arm Limited or its affiliates.
//
//            Release Information : MERCURY-r1p0-00rel0
//
//-----------------------------------------------------------------------------
// Verilog-2001 (IEEE Std 1364-2001)
//-----------------------------------------------------------------------------


`include "mercury_params.v"
`include "mercury_dpu_params.v"

module mercury_dpu_crypto_aes_mixcol (
  input  wire         aes_invmixcols_f3_i,
  input  wire  [31:0] crypto_a_data_f3_i,
  output wire  [31:0] aes_mixcol_data_f3_o
);

  wire [10:0] raw_mixcol_com [3:0];
  wire [10:0] raw_mixcol_inv [3:0];


  function [10:0] gf_mul_8x4;
    input [3:0] a;
    input [7:0] b;

    gf_mul_8x4 = {3'b000, ({8{a[0]}} & b)        } ^
                 {2'b00,  ({8{a[1]}} & b),   1'b0} ^
                 {1'b0,   ({8{a[2]}} & b),  2'b00} ^
                 {        ({8{a[3]}} & b), 3'b000};
  endfunction

  function [7:0] gf_reduce_11;
    input [10:0] a;

    gf_reduce_11 = a[7:0]
                   ^ ({8{a[ 8]}} & 8'h1b)
                   ^ ({8{a[ 9]}} & 8'h36)
                   ^ ({8{a[10]}} & 8'h6c);
  endfunction

  assign raw_mixcol_com[0] = gf_mul_8x4(4'd2, crypto_a_data_f3_i[ 7: 0]) ^
                             gf_mul_8x4(4'd3, crypto_a_data_f3_i[15: 8]) ^
                             gf_mul_8x4(4'd1, crypto_a_data_f3_i[23:16]) ^
                             gf_mul_8x4(4'd1, crypto_a_data_f3_i[31:24]);

  assign raw_mixcol_com[1] = gf_mul_8x4(4'd1, crypto_a_data_f3_i[ 7: 0]) ^
                             gf_mul_8x4(4'd2, crypto_a_data_f3_i[15: 8]) ^
                             gf_mul_8x4(4'd3, crypto_a_data_f3_i[23:16]) ^
                             gf_mul_8x4(4'd1, crypto_a_data_f3_i[31:24]);

  assign raw_mixcol_com[2] = gf_mul_8x4(4'd1, crypto_a_data_f3_i[ 7: 0]) ^
                             gf_mul_8x4(4'd1, crypto_a_data_f3_i[15: 8]) ^
                             gf_mul_8x4(4'd2, crypto_a_data_f3_i[23:16]) ^
                             gf_mul_8x4(4'd3, crypto_a_data_f3_i[31:24]);

  assign raw_mixcol_com[3] = gf_mul_8x4(4'd3, crypto_a_data_f3_i[ 7: 0]) ^
                             gf_mul_8x4(4'd1, crypto_a_data_f3_i[15: 8]) ^
                             gf_mul_8x4(4'd1, crypto_a_data_f3_i[23:16]) ^
                             gf_mul_8x4(4'd2, crypto_a_data_f3_i[31:24]);

  assign raw_mixcol_inv[0] = gf_mul_8x4(4'd12, crypto_a_data_f3_i[ 7: 0]) ^
                             gf_mul_8x4(4'd8,  crypto_a_data_f3_i[15: 8]) ^
                             gf_mul_8x4(4'd12, crypto_a_data_f3_i[23:16]) ^
                             gf_mul_8x4(4'd8,  crypto_a_data_f3_i[31:24]);

  assign raw_mixcol_inv[1] = gf_mul_8x4(4'd8,  crypto_a_data_f3_i[ 7: 0]) ^
                             gf_mul_8x4(4'd12, crypto_a_data_f3_i[15: 8]) ^
                             gf_mul_8x4(4'd8,  crypto_a_data_f3_i[23:16]) ^
                             gf_mul_8x4(4'd12, crypto_a_data_f3_i[31:24]);

  assign raw_mixcol_inv[2] = gf_mul_8x4(4'd12, crypto_a_data_f3_i[ 7: 0]) ^
                             gf_mul_8x4(4'd8,  crypto_a_data_f3_i[15: 8]) ^
                             gf_mul_8x4(4'd12, crypto_a_data_f3_i[23:16]) ^
                             gf_mul_8x4(4'd8,  crypto_a_data_f3_i[31:24]);

  assign raw_mixcol_inv[3] = gf_mul_8x4(4'd8,  crypto_a_data_f3_i[ 7: 0]) ^
                             gf_mul_8x4(4'd12, crypto_a_data_f3_i[15: 8]) ^
                             gf_mul_8x4(4'd8,  crypto_a_data_f3_i[23:16]) ^
                             gf_mul_8x4(4'd12, crypto_a_data_f3_i[31:24]);


  assign aes_mixcol_data_f3_o[ 7: 0] = gf_reduce_11(raw_mixcol_com[0] ^ ({11{aes_invmixcols_f3_i}} & raw_mixcol_inv[0]));
  assign aes_mixcol_data_f3_o[15: 8] = gf_reduce_11(raw_mixcol_com[1] ^ ({11{aes_invmixcols_f3_i}} & raw_mixcol_inv[1]));
  assign aes_mixcol_data_f3_o[23:16] = gf_reduce_11(raw_mixcol_com[2] ^ ({11{aes_invmixcols_f3_i}} & raw_mixcol_inv[2]));
  assign aes_mixcol_data_f3_o[31:24] = gf_reduce_11(raw_mixcol_com[3] ^ ({11{aes_invmixcols_f3_i}} & raw_mixcol_inv[3]));

endmodule

`define MERCURY_UNDEFINE
`include "mercury_params.v"
`include "mercury_dpu_params.v"
`undef MERCURY_UNDEFINE
