//-----------------------------------------------------------------------------
// The confidential and proprietary information contained in this file may
// only be used by a person authorised under and to the extent permitted
// by a subsisting licensing agreement from Arm Limited or its affiliates.
//
//            (C) COPYRIGHT 2012-2016 Arm Limited or its affiliates.
//                ALL RIGHTS RESERVED
//
// This entire notice must be reproduced on all copies of this file
// and copies of this file may only be made by a person if such person is
// permitted to do so under the terms of a subsisting license agreement
// from Arm Limited or its affiliates.
//
//            Release Information : MERCURY-r1p0-00rel0
//
//-----------------------------------------------------------------------------
// Verilog-2001 (IEEE Std 1364-2001)
//-----------------------------------------------------------------------------


`include "mercury_params.v"
`include "mercury_dpu_params.v"

module mercury_dpu_crypto_sha1cpm (
  input  wire [`MERCURY_CRYPTO_OP_W-1:0] crypto_op_i,
  input  wire                 [159:0] hash_data_i,
  input  wire                  [31:0] schedule_data_i,
  output wire                 [159:0] sha1cpm_res_o
);


  reg  [31: 0] t;


  wire [31: 0] a;
  wire [31: 0] a_prime;
  wire [31: 0] b;
  wire [31: 0] b_prime;
  wire [31: 0] c;
  wire [31: 0] c_prime;
  wire [31: 0] d;
  wire [31: 0] d_prime;
  wire [31: 0] e;
  wire [31: 0] e_prime;


  assign a = hash_data_i[ 31:  0];
  assign b = hash_data_i[ 63: 32];
  assign c = hash_data_i[ 95: 64];
  assign d = hash_data_i[127: 96];
  assign e = hash_data_i[159:128];

  always @*
    case (crypto_op_i)
      `MERCURY_CRYPTO_OP_SHA1C: t = (b & c) | (~b & d);
      `MERCURY_CRYPTO_OP_SHA1P: t = b ^ c ^ d;
      `MERCURY_CRYPTO_OP_SHA1M: t = (b & c) | (c & d) | (d & b);
      default:             t = {32{1'bx}};
    endcase

  assign a_prime = e + {a[26:0], a[31:27]} + t + schedule_data_i;
  assign b_prime = a;
  assign c_prime = {b[1:0], b[31:2]};
  assign d_prime = c;
  assign e_prime = d;
  assign sha1cpm_res_o = {e_prime, d_prime, c_prime, b_prime, a_prime};

endmodule

`define MERCURY_UNDEFINE
`include "mercury_params.v"
`include "mercury_dpu_params.v"
`undef MERCURY_UNDEFINE
