//-----------------------------------------------------------------------------
// The confidential and proprietary information contained in this file may
// only be used by a person authorised under and to the extent permitted
// by a subsisting licensing agreement from Arm Limited or its affiliates.
//
//            (C) COPYRIGHT 2012-2016 Arm Limited or its affiliates.
//                ALL RIGHTS RESERVED
//
// This entire notice must be reproduced on all copies of this file
// and copies of this file may only be made by a person if such person is
// permitted to do so under the terms of a subsisting license agreement
// from Arm Limited or its affiliates.
//
//            Release Information : MERCURY-r1p0-00rel0
//
//-----------------------------------------------------------------------------
// Verilog-2001 (IEEE Std 1364-2001)
//-----------------------------------------------------------------------------


`include "mercury_params.v"
`include "mercury_dpu_params.v"

module mercury_dpu_crypto_sha256su1 (
  input  wire  [63:0] op1_data_i,
  input  wire  [63:0] t0_data_i,
  input  wire  [63:0] t1_data_i,
  output wire  [63:0] sha256su1_res_o
);


  wire [31:0] elt0_step0;
  wire [31:0] elt0_step1;
  wire [31:0] elt1_step0;
  wire [31:0] elt1_step1;


  assign elt0_step0 = t1_data_i[31: 0];
  assign elt1_step0 = t1_data_i[63:32];

  assign elt0_step1 = {elt0_step0[16:0], elt0_step0[31:17]} ^ {elt0_step0[18:0], elt0_step0[31:19]} ^ {10'h000, elt0_step0[31:10]};
  assign elt1_step1 = {elt1_step0[16:0], elt1_step0[31:17]} ^ {elt1_step0[18:0], elt1_step0[31:19]} ^ {10'h000, elt1_step0[31:10]};

  assign sha256su1_res_o[31: 0] = elt0_step1 + op1_data_i[31: 0] + t0_data_i[31: 0];
  assign sha256su1_res_o[63:32] = elt1_step1 + op1_data_i[63:32] + t0_data_i[63:32];

endmodule

`define MERCURY_UNDEFINE
`include "mercury_params.v"
`include "mercury_dpu_params.v"
`undef MERCURY_UNDEFINE
